import abc
import typing
from collections import deque

from .event import Event

# If a new class inherits from any ABC, then the new class's metaclass has to
# inherit from ABCMeta. If a new class inherits from typing.Generic, and
# you're using Python 3.6, then the new class's metaclass has to
# inherit from typing.GenericMeta. Some of the classes that want to use Final
# or NoPublicConstructor inherit from ABCs and generics, so Final has to
# inherit from these metaclasses. Fortunately, GenericMeta inherits from
# ABCMeta, so inheriting from GenericMeta alone is sufficient (when it
# exists at all).

if not typing.TYPE_CHECKING and hasattr(typing, "GenericMeta"):
    BaseMeta = typing.GenericMeta
else:
    BaseMeta = abc.ABCMeta

T = typing.TypeVar("T")


class NoPublicConstructor(BaseMeta):
    """Metaclass that enforces a class to be final (i.e., subclass not allowed)
    and ensures a private constructor.

    If a class uses this metaclass like this::

        class SomeClass(metaclass=NoPublicConstructor):
            pass

    The metaclass will ensure that no sub class can be created, and that no instance
    can be initialized.

    If you try to instantiate your class (SomeClass()), a TypeError will be thrown.

    Raises
    ------
    - TypeError if a sub class or an instance is created.
    """

    def __call__(cls: typing.Type[T], *args: typing.Any, **kwargs: typing.Any) -> T:
        raise TypeError(
            f"{cls.__module__}.{cls.__qualname__} has no public constructor"
        )

    def _create(cls: typing.Type[T], *args: typing.Any, **kwargs: typing.Any) -> T:
        return super().__call__(*args, **kwargs)  # type: ignore


Id = typing.TypeVar("Id")


class Entity(typing.Generic[Id], metaclass=NoPublicConstructor):
    def __init__(self, *, unique_id: Id):
        self._unique_id = unique_id
        self._outbox: typing.Deque[Event] = deque()

    @property
    def id(self) -> Id:
        return self._unique_id

    @property
    def outbox(self) -> typing.Deque[Event]:
        return self._outbox

    def clear_outbox(self) -> typing.List[Event]:
        consumed_data = list(self._outbox)
        self._outbox.clear()
        return consumed_data

    def as_dictionary(self) -> typing.Dict[str, typing.Any]:
        return self.__dict__

    def __eq__(self, other: typing.Any) -> bool:
        if isinstance(other, self.__class__):
            return self.id == other.id
        return False



import abc
import datetime as dt
import enum
import uuid
from typing import (
    AsyncContextManager,
    AsyncIterable,
    Awaitable,
    Callable,
    Iterable,
    List,
    Optional,
)

from eventual.model import Entity, EventBody


class Message(abc.ABC):
    @property
    @abc.abstractmethod
    def body(self) -> EventBody:
        raise NotImplementedError

    @abc.abstractmethod
    def acknowledge(self) -> None:
        raise NotImplementedError

    def __repr__(self) -> str:
        return self.body["id"]


class MessageBroker(abc.ABC):
    @abc.abstractmethod
    async def send_event_body_stream(
        self, event_body_stream: AsyncIterable[EventBody]
    ) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    async def message_stream(self) -> AsyncIterable[Message]:
        raise NotImplementedError


class Guarantee(str, enum.Enum):
    NO_MORE_THAN_ONCE = "NO_MORE_THAN_ONCE"
    EXACTLY_ONCE = "EXACTLY_ONCE"
    AT_LEAST_ONCE = "AT_LEAST_ONCE"


class MessageDispatcher(abc.ABC):
    @abc.abstractmethod
    async def dispatch_from_stream(
        self, message_stream: AsyncIterable[Message]
    ) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    async def dispatch_from_exchange(self, message_exchange: MessageBroker) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def register(
        self,
        event_type_seq: List[str],
        fn: Callable[[Message], Awaitable[None]],
        guarantee: Guarantee,
        delay_on_exc: float,
    ) -> None:
        raise NotImplementedError

    def subscribe(
        self,
        event_type_seq: List[str],
        guarantee: Guarantee = Guarantee.AT_LEAST_ONCE,
        delay_on_exc: float = 1.0,
    ) -> Callable[
        [Callable[[Message], Awaitable[None]]], Callable[[Message], Awaitable[None]]
    ]:
        def decorator(
            fn: Callable[[Message], Awaitable[None]]
        ) -> Callable[[Message], Awaitable[None]]:
            self.register(event_type_seq, fn, guarantee, delay_on_exc)
            return fn

        return decorator


class EventStore(abc.ABC):
    @abc.abstractmethod
    def clear_outbox_atomically(self, *entity_seq: Entity) -> AsyncContextManager[None]:
        raise NotImplementedError

    async def clear_outbox(self, entity_seq: Iterable[Entity]) -> None:
        for entity in entity_seq:
            # Make a copy because during asynchronous processing
            # someone can add messages to the outbox.
            event_seq = entity.clear_outbox()
            # One may think that the order, in which the events are written here,
            # is important for sourcing the events later. In reality every event has a timestamp, which dictates
            # its position in the sequence.
            for event in event_seq:
                await self.schedule_event_out(
                    event_id=event.id, body=event.encode_body()
                )
            if entity.outbox:
                raise ValueError("writing to outbox after clearing loses events")

    @abc.abstractmethod
    async def schedule_event_out(
        self,
        event_id: uuid.UUID,
        body: EventBody,
        send_after: Optional[dt.datetime] = None,
    ) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    async def is_event_handled(self, event_id: uuid.UUID) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    async def mark_event_handled(
        self,
        event_body: EventBody,
        guarantee: Guarantee,
    ) -> uuid.UUID:
        raise NotImplementedError

    @abc.abstractmethod
    async def mark_event_dispatched(self, event_body: EventBody) -> uuid.UUID:
        raise NotImplementedError

    @abc.abstractmethod
    def handle_exactly_once(self, message: Message) -> AsyncContextManager[EventBody]:
        raise NotImplementedError

    @abc.abstractmethod
    def handle_no_more_than_once(
        self, message: Message
    ) -> AsyncContextManager[EventBody]:
        raise NotImplementedError

    @abc.abstractmethod
    def handle_at_least_once(self, message: Message) -> AsyncContextManager[EventBody]:
        raise NotImplementedError

    @abc.abstractmethod
    def event_body_stream(self) -> AsyncIterable[EventBody]:
        raise NotImplementedError



import abc
import uuid
from typing import Generic, Iterable, Optional, TypeVar

from eventual.model import Entity

M = TypeVar("M", bound=Entity)


class RepositoryAllMixin(Generic[M], abc.ABC):
    @abc.abstractmethod
    async def all(self) -> Iterable[M]:
        raise NotImplementedError


class RepositoryCountMixin(Generic[M], abc.ABC):
    @abc.abstractmethod
    async def count(self) -> int:
        raise NotImplementedError


class RepositoryGetMixin(Generic[M], abc.ABC):
    @abc.abstractmethod
    async def get(self, unique_id: uuid.UUID) -> Optional[M]:
        raise NotImplementedError


class RepositoryCreateMixin(Generic[M], abc.ABC):
    @abc.abstractmethod
    async def create(self, value: M) -> M:
        raise NotImplementedError


class RepositoryUpdateMixin(Generic[M], abc.ABC):
    @abc.abstractmethod
    async def update(self, value: M) -> Optional[M]:
        raise NotImplementedError


class RepositoryDeleteMixin(Generic[M], abc.ABC):
    async def delete(self, value: M) -> Optional[M]:
        return await self.delete_by_id(value.id)

    @abc.abstractmethod
    async def delete_by_id(self, unique_id: uuid.UUID) -> Optional[M]:
        raise NotImplementedError


class RepositoryMixin(
    RepositoryAllMixin[M],
    RepositoryCountMixin[M],
    RepositoryGetMixin[M],
    RepositoryCreateMixin[M],
    RepositoryUpdateMixin[M],
    RepositoryDeleteMixin[M],
    abc.ABC,
):
    pass



import abc
from typing import AsyncContextManager, Callable


class InterruptWork(Exception):
    pass


class WorkUnit(abc.ABC):
    def __init__(self) -> None:
        self.was_committed = False

    @abc.abstractmethod
    async def rollback(self) -> None:
        raise NotImplementedError


WorkUnitFactory = Callable[[], AsyncContextManager[WorkUnit]]
